import json

import click

from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import API_GATEWAY_TYPE
from syndicate.core.generators import (_read_content_from_file,
                                       _write_content_to_file)
from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator, BaseConfigurationGenerator

_LOG = get_logger(
    'syndicate.core.generators.deployment_resources.api_gateway_generator')
USER_LOG = get_user_logger()


class ApiGatewayGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = API_GATEWAY_TYPE
    CONFIGURATION = {
        'deploy_stage': None,
        'dependencies': list,
        'resources': dict,
        'minimum_compression_size': None
    }


class ApiGatewayConfigurationGenerator(BaseConfigurationGenerator):
    """Contains common method for resource and method generators"""

    def __init__(self, **kwargs):
        self.api_gateway_name = kwargs.pop('api_name')
        super().__init__(**kwargs)

    def get_meta_with_resource(self, paths_with_api: list, resource_path: str):
        """Returns the path to deployment resources with api which contains
        resource with given path"""
        for path in paths_with_api:
            deployment_resources = json.loads(_read_content_from_file(path))
            if resource_path in deployment_resources[self.api_gateway_name][
                'resources']:
                _LOG.info(f"Found resource '{resource_path}' in api "
                          f"'{self.api_gateway_name}' in file '{path}'")
                return path, deployment_resources


class ApiGatewayResourceGenerator(ApiGatewayConfigurationGenerator):
    CONFIGURATION = {
        'enable_cors': bool
    }

    def __init__(self, **kwargs):
        self.resource_path = kwargs.pop('path')
        super().__init__(**kwargs)

    def write(self):
        """Adds resource to API gateway"""
        paths_with_api = self._get_resource_meta_paths(self.api_gateway_name,
                                                       API_GATEWAY_TYPE)
        if not paths_with_api:
            message = f"Api gateway '{self.api_gateway_name}' was not found"
            _LOG.error(message)
            raise ValueError(message)
        USER_LOG.info(f"Adding resource '{self.resource_path}' to api "
                      f"'{self.api_gateway_name}'...")

        path_with_resources = self.get_meta_with_resource(paths_with_api,
                                                          self.resource_path)
        if not path_with_resources:
            path_with_api = paths_with_api[0]
            deployment_resources = json.loads(
                _read_content_from_file(path_with_api))
        else:
            path_with_api, deployment_resources = path_with_resources
            message = f"Resource '{self.resource_path}' was found in " \
                      f"api gateway '{self.api_gateway_name}' in file " \
                      f"'{path_with_api}'"
            _LOG.warning(f"Found duplicate while generating meta. "
                         f"{message}")
            if not click.confirm(f"{message} Overwrite?"):
                USER_LOG.warning(f"Skipping resource "
                                 f"'{self.resource_path}'")
                raise RuntimeError

        USER_LOG.info(f"Adding resource '{self.resource_path}' to api "
                      f"'{self.api_gateway_name}'...")
        deployment_resources[self.api_gateway_name]['resources'][
            self.resource_path] = self._resolve_configuration()
        _write_content_to_file(path_with_api,
                               json.dumps(deployment_resources, indent=2))


class ApiGatewayResourceMethodGenerator(ApiGatewayConfigurationGenerator):
    CONFIGURATION = {
        'authorization_type': "NONE",
        'integration_type': 'mock',
        'lambda_name': None,
        'lamdba_region': None,
        'api_key_required': bool,
        'method_request_parameters': dict,
        'integration_request_body_template': dict,
        'responses': list,
        'integration_responses': list,
        'default_error_pattern': True,
    }

    def __init__(self, **kwargs):
        self.resource_path = kwargs.pop('path')
        self.method = kwargs.pop('method')
        super().__init__(**kwargs)

    def write(self):
        paths_with_api = self._get_resource_meta_paths(self.api_gateway_name,
                                                       API_GATEWAY_TYPE)
        if not paths_with_api:
            message = f"Api gateway '{self.api_gateway_name}' was not found"
            _LOG.error(message)
            raise ValueError(message)

        path_with_resources = self.get_meta_with_resource(paths_with_api,
                                                          self.resource_path)
        if not path_with_resources:
            message = f"Resouce '{self.resource_path}' wasn't found in api " \
                      f"'{self.api_gateway_name}'"
            _LOG.error(message)
            raise ValueError(message)

        path_with_api, deployment_resources = path_with_resources

        if self.method in deployment_resources[self.api_gateway_name][
            'resources'][self.resource_path]:
            message = f"Method '{self.method}' was found in resource " \
                      f"'{self.resource_path}' from gateway " \
                      f"'{self.api_gateway_name}' in file '{path_with_api}'"
            _LOG.warning(f"Found duplicate while generating meta. "
                         f"{message}")
            if not click.confirm(f"{message} Overwrite?"):
                USER_LOG.warning(f"Skipping resource "
                                 f"'{self.resource_path}'")
                raise RuntimeError

        USER_LOG.info(f"Adding method '{self.method}' to resource "
                      f"'{self.resource_path}' to api "
                      f"'{self.api_gateway_name}'...")

        deployment_resources[self.api_gateway_name]['resources'][
            self.resource_path][self.method] = \
            self._resolve_configuration()
        _write_content_to_file(path_with_api,
                               json.dumps(deployment_resources, indent=2))

    def _resolve_configuration(self, defaults_dict=None):
        if self._dict.get('integration_type') == 'lambda':
            self.validate_integration_lambda_existence()
        else:
            if 'lamdba_name' in self._dict:
                self._dict.pop('lambda_name')
        return super()._resolve_configuration()

    def validate_integration_lambda_existence(self):
        from syndicate.core import PROJECT_STATE
        lambda_name = self._dict.get('lambda_name')
        _LOG.info(f"Validating existence of lambda: {lambda_name}")

        if not lambda_name in PROJECT_STATE.lambdas:
            message = f"Lambda '{lambda_name}' wasn't found"
            _LOG.error(f"Validation error: {message}")
            raise ValueError(message)
        _LOG.info(f"Validation successfully finished, lambda exists")



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os

import click
from syndicate.core.conf.generator import generate_configuration_files
from syndicate.core.generators.lambda_function import (
    generate_lambda_function)
from syndicate.core.generators.project import (generate_project_structure,
                                               PROJECT_PROCESSORS)
from syndicate.core.groups.meta import meta
from syndicate.core.helper import (timeit, OrderedGroup,
                                   check_bundle_bucket_name,
                                   check_prefix_suffix_length,
                                   resolve_project_path,
                                   check_lambdas_names, DictParamType)

GENERATE_GROUP_NAME = 'generate'
GENERATE_PROJECT_COMMAND_NAME = 'project'
GENERATE_CONFIG_COMMAND_NAME = 'config'
PROJECT_PATH_HELP = 'Path to project folder. ' \
                    'Default value: current working directory'


@click.group(name=GENERATE_GROUP_NAME, cls=OrderedGroup)
def generate():
    """Generates project, lambda or configs"""


@generate.command(name=GENERATE_PROJECT_COMMAND_NAME)
@click.option('--name', nargs=1, required=True, help='The project name')
@click.option('--path', nargs=1,
              help=PROJECT_PATH_HELP)
@timeit()
def project(name, path):
    """
    Generates project with all the necessary components and in a right
    folders/files hierarchy to start developing in a min.
    """
    click.echo(f'Project name: {name}')

    proj_path = os.getcwd() if not path else path
    if not os.access(proj_path, os.X_OK | os.W_OK):
        click.echo(
            f"Incorrect permissions for the provided path '{proj_path}'")
        return
    click.echo(f'Project path: {proj_path}')
    generate_project_structure(project_name=name,
                               project_path=proj_path)


@generate.command(name='lambda')
@click.option('--name', multiple=True, type=str,
              required=True, callback=check_lambdas_names,
              help='(multiple) The lambda function name')
@click.option('--runtime', required=True,
              help='Lambda\'s runtime. If multiple lambda names are specified,'
                   ' the runtime will be applied to all lambdas',
              type=click.Choice(PROJECT_PROCESSORS))
@click.option('--project_path', nargs=1,
              help="Path to the project folder. Default value: the one "
                   "from the current config if it exists. "
                   "Otherwise - the current working directory",
              callback=resolve_project_path)
@timeit()
def lambda_function(name, runtime, project_path):
    """
    Generates required environment for lambda function
    """
    if not os.access(project_path, os.F_OK):
        click.echo(f"The provided path {project_path} doesn't exist")
        return
    elif not os.access(project_path, os.W_OK) or not os.access(project_path,
                                                               os.X_OK):
        click.echo(f"Incorrect permissions for the provided path "
                   f"'{project_path}'")
        return
    click.echo(f'Lambda names: {name}')
    click.echo(f'Runtime: {runtime}')
    click.echo(f'Project path: {project_path}')
    generate_lambda_function(project_path=project_path,
                             runtime=runtime,
                             lambda_names=name)


@generate.command(name=GENERATE_CONFIG_COMMAND_NAME)
@click.option('--name',
              required=True,
              help='Name of the configuration to create. '
                   'Generated config will be created in folder '
                   '.syndicate-config-{name}. May contain name '
                   'of the environment')
@click.option('--region',
              help='The region that is used to deploy the application',
              required=True)
@click.option('--bundle_bucket_name',
              help='Name of the bucket that is used for uploading artifacts.'
                   ' It will be created if specified.', required=True,
              callback=check_bundle_bucket_name)
@click.option('--access_key',
              help='AWS access key id that is used to deploy the application. '
                   'Retrieved from session by default')
@click.option('--secret_key',
              help='AWS secret key that is used to deploy the application. '
                   'Retrieved from session by default')
@click.option('--config_path',
              help='Path to store generated configuration file')
@click.option('--project_path',
              help=PROJECT_PATH_HELP)
@click.option('--prefix',
              help='Prefix that is added to project names while deployment '
                   'by pattern: {prefix}resource_name{suffix}. '
                   'Must be less than or equal to 5',
              callback=check_prefix_suffix_length)
@click.option('--suffix',
              help='Suffix that is added to project names while deployment '
                   'by pattern: {prefix}resource_name{suffix}. '
                   'Must be less than or equal to 5',
              callback=check_prefix_suffix_length)
@click.option('--use_temp_creds', type=bool, default=False,
              help='Indicates Syndicate to generate and use temporary AWS '
                   'credentials')
@click.option('--access_role', type=str,
              help='Indicates Syndicate to use this role\'s temporary AWS '
                   'credentials. Cannot be used if \'--use_temp_creds\' is '
                   'equal to true')
@click.option('--serial_number', type=str,
              help='The identification number of the MFA device that is '
                   'associated with the IAM user which will be used for '
                   'deployment. If specified MFA token will be asked before '
                   'making actions')
@click.option('--tags', type=DictParamType(),
              help='Tags to add to the config. They will be added to all the '
                   'resources during deployment')
@click.option('--iam_permissions_boundary', type=str,
              help='Common permissions boundary arn to add to all the roles')
@timeit()
def config(name, config_path, project_path, region, access_key, secret_key,
           bundle_bucket_name, prefix, suffix, use_temp_creds, access_role,
           serial_number, tags, iam_permissions_boundary):
    """
    Creates Syndicate configuration files
    """
    generate_configuration_files(name=name,
                                 config_path=config_path,
                                 project_path=project_path,
                                 region=region,
                                 access_key=access_key,
                                 secret_key=secret_key,
                                 bundle_bucket_name=bundle_bucket_name,
                                 prefix=prefix,
                                 suffix=suffix,
                                 use_temp_creds=use_temp_creds,
                                 access_role=access_role,
                                 serial_number=serial_number,
                                 tags=tags,
                                 iam_permissions_boundary=iam_permissions_boundary)


generate.add_command(meta)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import ipaddress
import re
import string
from typing import Optional

from syndicate.commons.log_helper import get_logger
from syndicate.core import ClientError
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj, chunks

_LOG = get_logger('syndicate.core.resources.s3_resource')


def validate_bucket_name(bucket_name: str):
    """Checks whether the given bucket name is valid.
    If the given name isn't valid, ValueError with an appropriate message
    is raised.
    :type bucket_name: str
    :param bucket_name: the name to check
    """
    bucket_name = bucket_name.strip()
    _LOG.info(f"Starting validating bucket name '{bucket_name}'")
    error = None
    if not bucket_name or not 3 <= len(bucket_name) <= 63:
        error = 'Bucket name must be between 3 and 63 characters long'
    else:
        invalid_characters = re.search('[^a-z0-9.-]', bucket_name)
        if invalid_characters:
            character = invalid_characters.group()
            if character in string.ascii_uppercase:
                error = 'Bucket name must not contain uppercase characters'
            else:
                error = f'Bucket name contains invalid characters: {character}'
        elif any(bucket_name.startswith(symbol) for symbol in '.-'):
            error = 'Bucket name must start with a lowercase letter or number'
        elif any(bucket_name.endswith(symbol) for symbol in '.-'):
            error = 'Bucket name must not end with dash or period'
        elif '..' in bucket_name:
            error = 'Bucket name must not contain two adjacent periods'
        elif '.-' in bucket_name or '-.' in bucket_name:
            error = 'Bucket name must not contain dash next to period'
        else:
            try:
                ipaddress.ip_address(bucket_name)
                error = 'Bucket name must not resemble an IP address'
            except ValueError:
                pass
    if error:
        _LOG.warning(error)
        raise ValueError(error)
    _LOG.info(f"Finished validating bucket name '{bucket_name}'")


class S3Resource(BaseResource):

    def __init__(self, s3_conn, account_id) -> None:
        self.s3_conn = s3_conn
        self.account_id = account_id

    def create_s3_bucket(self, args):
        return self.create_pool(self._create_s3_bucket_from_meta, args)

    def describe_bucket(self, name, meta):
        arn = self.get_bucket_arn(name)
        acl_response = self.s3_conn.get_bucket_acl(name)
        location_response = self.s3_conn.get_bucket_location(name)
        bucket_policy = self.s3_conn.get_bucket_policy(name)
        response = {
            'bucket_acl': acl_response,
            'location': location_response,
        }
        if bucket_policy:
            response['policy'] = bucket_policy
        return {
            arn: build_description_obj(response, name, meta)
        }

    @staticmethod
    def get_bucket_arn(name):
        return 'arn:aws:s3:::{0}'.format(name)

    @unpack_kwargs
    def _create_s3_bucket_from_meta(self, name, meta):
        if self.s3_conn.is_bucket_exists(name):
            _LOG.warn('{0} bucket exists.'.format(name))
            return self.describe_bucket(name, meta)
        self.s3_conn.create_bucket(name, meta.get('acl'), meta.get('location'))
        policy = meta.get('policy')
        if policy:
            self.s3_conn.add_bucket_policy(name, policy)
            _LOG.debug('Policy on {0} S3 bucket is set up.'.format(name))
        _LOG.info('Created S3 bucket {0}.'.format(name))
        rules = meta.get('LifecycleConfiguration')
        if rules:
            self.s3_conn.add_bucket_rule(name, rules)
            _LOG.debug('Rules on {0} S3 bucket are set up.'.format(name))
        cors_configuration = meta.get('cors')
        if cors_configuration:
            self.s3_conn.put_cors(bucket_name=name, rules=cors_configuration)
        public_access_block = meta.get('public_access_block', {})
        if not all([isinstance(param, bool) for param in
                    public_access_block.values()]):
            message = f'Parameters inside public_access_block should have ' \
                      f'bool type'
            _LOG.error(message)
            raise AssertionError(message)
        self.s3_conn.put_public_access_block(name,
                                             **public_access_block)
        return self.describe_bucket(name, meta)

    def _delete_objects(self, bucket_name, keys):
        response = self.s3_conn.delete_objects(bucket_name, keys)
        errors = response.get('Errors')
        if errors:
            error_keys = [{
                'Key': i['Key'],
                'VersionId': i['VersionId']
            } for i in errors]
            return error_keys
        else:
            return []

    def remove_buckets(self, args):
        self.create_pool(self._remove_bucket, args)

    @unpack_kwargs
    def _remove_bucket(self, arn, config):
        bucket_name = config['resource_name']
        try:
            errors = []
            keys = self.s3_conn.list_object_versions(bucket_name)
            if keys:
                for s3_keys in chunks(keys, 1000):
                    errors.extend(self._delete_objects(bucket_name, s3_keys))

            markers = self.s3_conn.list_object_markers(bucket_name)
            if markers:
                for s3_markers in chunks(markers, 1000):
                    errors.extend(
                        self._delete_objects(bucket_name, s3_markers))

            if errors:
                raise AssertionError('Error occurred while deleting S3 objects'
                                     ' from {0} bucket. Not deleted keys: '
                                     '{1}'.format(bucket_name, str(errors)))
            else:
                self.s3_conn.delete_bucket(bucket_name)
                _LOG.info('S3 bucket {0} was removed.'.format(bucket_name))
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchBucket':
                _LOG.warn('S3 bucket {0} is not found'.format(bucket_name))
            else:
                raise e

    def build_bucket_arn(self, maybe_arn: str) -> Optional[str]:
        if not isinstance(maybe_arn, str):
            return
        if self.is_bucket_arn(maybe_arn):
            return maybe_arn
        return f'arn:aws:s3:::{maybe_arn}'

    @staticmethod
    def is_bucket_arn(maybe_arn: str) -> bool:
        # TODO add files keys support to regex
        return bool(re.match(r'^arn:aws:s3:::[a-z0-9.-]{3,63}/?$',
                             maybe_arn))



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import S3_BUCKET_TYPE


class S3Generator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = S3_BUCKET_TYPE
    CONFIGURATION = {
        'acl': 'private',
        'location': None,
        'cors': list,
        'policy': dict,
        'public_access_block': {
            'block_public_acls': True,
            'ignore_public_acls': True,
            'block_public_policy': True,
            'restrict_public_buckets': True
        }
    }
